"""
    FunctionTests driver module.
"""

import sys, getopt, re, ConfigParser

_marker = []

class FTRunner:

    _verbosity              = 1
    _time_requests          = 1
    _check_responses        = 1
    _check_elapsed_times    = 1
    _time_requests          = 1
    _config_file            = '.zftrc'
    _portal_url             = 'http://localhost:8080/'
    _site_path              = '/'
    _filenames              = ()
    _app                    = _marker
    
    def __init__( self, args ):
        self._defaults = {}
        self.parseOptions( args )

    def printUsage( self, message=None ):
        """
            Print command-line help.
        """
        print """\

FTRunner [-htTvq] [-c config_file] \\
          [-u portal_url] [-r site_path] \\
          [name1=value]* scenario_1 ...

    -?, --help                      Print this help message

    -t, --time-requests             Time requests (default)

    -T, --no-time-requests          Don't time requests

    -r, --check-responses           Check expected HTTP respose codes (default)

    -R, --no-check-responses        Don't check expected HTTP respose codes

    -e, --check-elapsed_times       Check elapsed times (default)

    -E, --no-check-elapsed_times    Don't check elapsed times

    -v, --verbose                   Increment verbosity (default is '1')

    -q, --quiet                     Set verbosity to '0'

    -c, --config-file               Use defaults from 'config_file'
                                    (default is '%(CONFIG_FILE)s')

    -u, --portal-url                Use 'portal_url' as test target server
                                    (default is '%(PORTAL_URL)s')

    -p, --site-path                 Use 'site_path' as test target path
                                    (default is '%(SITE_PATH)s')

Notes:

- Actually running a test may require being able to 'import Zope',
  which means that the appropriate SOFTWARE_HOME, INSTANCE_HOME, and
  PYTHONPATH variables should be set up by the script which calls us.

%(MESSAGE)s
        """ % ( { 'CONFIG_FILE' : self._config_file
                , 'PORTAL_URL'  : self._portal_url
                , 'SITE_PATH'   : self._site_path
                , 'MESSAGE'     : message and 'Error: %s' % message or ''
                } )
        sys.exit( 1 )

    def parseOptions( self, args ):
        """
            Parse command line options.
        """
        verbosity = 1
        portal_url = site_path = config_file = None
        time_requests = check_responses = check_elapsed_times = None

        try:
            opts, arglist = getopt.getopt( args
                                           , "?tTrReEvqc:u:p:"
                                           , [ 'help'
                                             , 'time-requests'
                                             , 'no-time-requests'
                                             , 'check-responses'
                                             , 'no-check-responses'
                                             , 'check-elapsed_times'
                                             , 'no-check-elapsed_times'
                                             , 'verbose'
                                             , 'quiet'
                                             , 'config-file='
                                             , 'portal-url='
                                             , 'site-path='
                                             ]
                                           )
        except getopt.GetoptError, message:
            self.printUsage( message=message)
            
        for o, v in opts:

            if o == '-?' or o == '--help':
                self.printUsage()

            if o == '-t' or o == '--time-requests':
                time_requests = 1

            if o == '-T' or o == '--no-time-requests':
                time_requests = 0

            if o == '-r' or o == '--check-responses':
                check_responses = 1

            if o == '-R' or o == '--no-check-responses':
                check_responses = 0

            if o == '-e' or o == '--check-elapsed_times':
                check_elapsed_times = 1

            if o == '-E' or o == '--no-check-elapsed_times':
                check_elapsed_times = 0

            if o == '-v' or o == '--verbose':
                verbosity = verbosity + 1

            if o == '-q' or o == '--quiet':
                verbosity = 0

            if o == '-c' or o == '--config-file':
                config_file = v

            if o == '-u' or o == '--portal-url':
                portal_url = v

            if o == '-p' or o == '--site-path':
                site_path = v

        defaults, filenames = self.parseDefaults( arglist )

        if len( filenames ) == 0:
            self.printUsage( message='must specify at least one scenario' )

        self._filenames = filenames

        # First initial defaults, ...
        self._defaults = { 'portal_url' : self._portal_url
                         , 'site_path'  : self._site_path
                         }

        self._defaults.update( defaults )

        # ... then config file, ...
        if config_file is None:
            try:
                f = open( '.zftrc' )
            except IOError:
                pass
            else:
                config_file = '.zftrc'
        
        self._config_file = config_file

        if self._config_file is not None:
            self.readConfigFile( self._defaults )

        # ... then command-line args.
        if verbosity is not None:
            self._verbosity = verbosity

        if time_requests is not None:
            self._time_requests = time_requests

        if check_responses is not None:
            self._check_responses = check_responses

        if check_elapsed_times is not None:
            self._check_elapsed_times = check_elapsed_times

        if portal_url is not None:
            self._defaults[ 'portal_url' ] = portal_url

        if site_path is not None:
            self._defaults[ 'site_path' ] = site_path

    def parseDefaults( self, arglist ):
        """
            Return a dictionary containing name-value pairs from
            arglist (those whose arg matched 'name=value'), plus
            a sequence of "normal" arguments.
        """
        ARG_PATTERN = re.compile( r'([A-Za-z_][\w_]*)[ ]*=[ ]*(.*)' )

        defaults = {}
        filenames = []

        for arg in arglist:

            match = ARG_PATTERN.match( arg )

            if match:
                defaults[ match.group(1) ] = match.group(2)
            else:
                filenames.append( arg )

        return defaults, filenames

    def readConfigFile( self, defaults ):
        """
            Initialize defaults from the command line.
        """
        cp = ConfigParser.ConfigParser( defaults )
        try:
            f = open( self._config_file )
            cp.readfp( f )
            self._defaults.update( cp.defaults() )

            if 'Options' in cp.sections():

                self._time_requests = cp.get( 'Options'
                                            , 'Time_Requests'
                                            , self._time_requests
                                            )

                self._check_responses = cp.get( 'Options'
                                              , 'Check_Responses'
                                              , self._check_responses
                                              )

                self._check_elapsed_times = cp.get( 'Options'
                                                  , 'Check_Elapsed_Times'
                                                  , self._check_elapsed_times
                                                  )

                self._verbosity = cp.get( 'Options'
                                            , 'Verbosity'
                                            , self._verbosity
                                            )

        except IOError:
            print '*** No such config file: %s' % self._config_file
        except ConfigParser.Error, msg:
            print '*** Config file error: %s' % msg

    def printHeader( self, title ):
        print "=" * 79
        print "Test: %s" % title
        print "=" * 79
        print "%-12s: %-55s: %s" % ( 'Request', 'URL', 'Time (s)' )
        print "-" * 79

    def printInvocation( self, invocation ):
        """
            Print out a single invocation.
        """
        request = invocation.getRequest()

        print "%-12s: %-55s" % ( request.getName()[:12]
                            , request.getURI()[:55]
                            ),  # note comma

        print ' %0.2f' % invocation.getInterval()


    def printResult( self, result ):
        """
            Print out a Result.
        """
        if self._verbosity > 0:
            for invocation in result.listInvocations():
                self.printInvocation( invocation )

        if self._verbosity > 0:
            print "-" * 79
        outcome = result() and "OK" or "Failed"
        print "\n  Outcome: %s" % outcome

        errors = result.listErrors()
        if errors:
            print "\n  Errors"
        for error in errors:
            print error

        print

        if self._verbosity > 1:
            for dump in result.listDumpedResults():
                print dump
                print

    def getZopeApp( self ):
        """
            Get and return the Zope Application object, if available.
        """
        if self._app is _marker: # only try once

            try:
                import Zope
                self._app = Zope.app()
            except:
                self._app = None

        return self._app
        

    def runTestFile( self, filename ):
        """
            Run the scenario contained in 'filename'.
        """

        import Framework
        test = Framework.buildTest( filename, self._defaults )

        if self._verbosity > 0:
            self.printHeader( test.getTitle() )

        result = test( defaults=self._defaults
                     , app=self.getZopeApp()
                     , time_requests=self._time_requests
                     , check_responses=self._check_responses
                     , check_elapsed_times=self._check_elapsed_times
                     )

        self.printResult( result )

    def runAllTests( self ):

        for filename in self._filenames:
            self.runTestFile( filename )
    

if __name__ == '__main__':

    runner = FTRunner( sys.argv[1:] )
    runner.runAllTests()
